use crate::time::{format_time, Time};
use std::{borrow::Cow, fmt::Write as _};

pub(crate) enum Filter {
	/// Trim audio and video alike
	Trim {
		input: Cow<'static, str>,
		start: Option<Time>,
		duration: Option<Time>,
		output: Cow<'static, str>
	},

	/// Apply an alpha channel on the video. No audio.
	Alpha {
		input: Cow<'static, str>,
		alpha: f32,
		output: Cow<'static, str>
	},

	/// Overlay the one video over the other. The audio is copied.
	Overlay {
		video_input: Cow<'static, str>,
		overlay_input: Cow<'static, str>,
		x: Cow<'static, str>,
		y: Cow<'static, str>,
		output: Cow<'static, str>
	},

	/// Concatenate audio and video.
	Concat {
		inputs: Vec<Cow<'static, str>>,
		n: usize,
		output: Cow<'static, str>
	},

	/// Fade audio and video.
	Fade {
		input: Cow<'static, str>,
		direction: &'static str,
		start: Time,
		duration: Time,
		output: Cow<'static, str>
	},

	/// Generate silence. The video is copied.
	GenerateSilence {
		video: Cow<'static, str>,
		output: Cow<'static, str>
	},

	/// Fast forward. Too complex to explain. Its magic.
	FastForward {
		input: Cow<'static, str>,
		ffinput: Cow<'static, str>,
		start: Time,
		duration: Time,
		multiplier: usize,
		output: Cow<'static, str>
	}
}

impl Filter {
	pub(crate) fn is_video_filter(&self) -> bool {
		matches!(
			self,
			Self::Trim { .. }
				| Self::Alpha { .. }
				| Self::Concat { .. }
				| Self::Fade { .. }
				| Self::Overlay { .. }
		)
	}

	pub(crate) fn is_audio_filter(&self) -> bool {
		matches!(
			self,
			Self::Trim { .. }
				| Self::Concat { .. }
				| Self::Fade { .. }
				| Self::GenerateSilence { .. }
		)
	}

	pub(crate) fn append_to_complex_filter(
		&self,
		complex: &mut String,
		filter_idx: &mut usize
	) {
		match self {
			Self::Trim {
				input,
				start,
				duration,
				output
			} => {
				let mut args = String::new();
				if let Some(start) = start {
					write!(args, "start={start}");
				}
				if let Some(duration) = duration {
					if !args.is_empty() {
						args += ":";
					}
					write!(args, "duration={duration}");
				}
				writeln!(
					complex,
					"{}trim={args},setpts=PTS-STARTPTS{};",
					channel('v', input),
					channel('v', output)
				);
				writeln!(
					complex,
					"{}atrim={args},asetpts=PTS-STARTPTS{};",
					channel('a', input),
					channel('a', output)
				);
			},

			Self::Alpha {
				input,
				alpha,
				output
			} => {
				writeln!(
					complex,
					"{}format=yuva444p,colorchannelmixer=aa={alpha}{};",
					channel('v', input),
					channel('v', output)
				);
			},

			Self::Overlay {
				video_input,
				overlay_input,
				x,
				y,
				output
			} => {
				writeln!(
					complex,
					"{}{}overlay=x={x}:y={y}{};",
					channel('v', video_input),
					channel('v', overlay_input),
					channel('v', output)
				);
				writeln!(
					complex,
					"{}anull{};",
					channel('a', video_input),
					channel('a', output)
				);
			},

			Self::Concat { inputs, n, output } => {
				for i in inputs {
					write!(complex, "{}{}", channel('v', i), channel('a', i));
				}
				writeln!(
					complex,
					"concat=n={n}:v=1:a=1{}{};",
					channel('v', output),
					channel('a', output)
				);
			},

			Self::Fade {
				input,
				direction,
				start,
				duration,
				output
			} => {
				let args = format!("{direction}:st={start}:d={duration}");
				writeln!(
					complex,
					"{}fade={args}{};",
					channel('v', input),
					channel('v', output)
				);
				writeln!(
					complex,
					"{}afade=t={args}{};",
					channel('a', input),
					channel('a', output)
				);
			},

			Self::GenerateSilence { video, output } => {
				writeln!(
					complex,
					"{}null{};",
					channel('v', video),
					channel('v', output)
				);
				writeln!(complex, "aevalsrc=0:s=48000{};", channel('a', output));
			},

			Self::FastForward {
				input,
				ffinput,
				start,
				duration,
				multiplier,
				output
			} => {
				let end = *start + *duration;

				// ok so let's start by duplicating the audio and video 3 times
				let vin = next_tmp_3(filter_idx);
				let ain = next_tmp_3(filter_idx);
				writeln!(
					complex,
					"{}split=3{}{}{};",
					channel('v', input),
					vin[0],
					vin[1],
					vin[2]
				);
				writeln!(
					complex,
					"{}asplit=3{}{}{};",
					channel('a', input),
					ain[0],
					ain[1],
					ain[2]
				);

				// next we cut those audio/videos into before, ff, after
				let vcut = next_tmp_3(filter_idx);
				let acut = next_tmp_3(filter_idx);
				writeln!(complex, "{}trim=duration={start}{};", vin[0], vcut[0]);
				writeln!(complex, "{}atrim=duration={start}{};", ain[0], acut[0]);
				writeln!(
					complex,
					"{}trim=start={start}:duration={duration},setpts=PTS-STARTPTS{};",
					vin[1], vcut[1]
				);
				writeln!(
					complex,
					"{}atrim=start={start}:duration={duration},asetpts=PTS-STARTPTS{};",
					ain[1], acut[1]
				);
				writeln!(
					complex,
					"{}trim=start={end},setpts=PTS-STARTPTS{};",
					vin[2], vcut[2]
				);
				writeln!(
					complex,
					"{}atrim=start={end},asetpts=PTS-STARTPTS{};",
					ain[2], acut[2]
				);

				// now we speed up the ff part
				let vff = next_tmp(filter_idx);
				let aff = next_tmp(filter_idx);
				writeln!(complex, "{}setpts=PTS/{multiplier}{vff};", vcut[1]);
				writeln!(complex, "{}atempo={multiplier}{aff};", acut[1]);

				// and we overlay the vff part
				let voverlay = next_tmp(filter_idx);
				writeln!(
					complex,
					"{vff}{}overlay=x=main_w/2-overlay_w/2:y=main_h/2-overlay_h/2{voverlay};",
					channel('v', ffinput)
				);

				// and finally we concatenate everything back together
				writeln!(
					complex,
					"{}{}{voverlay}{aff}{}{}concat=n=3:v=1:a=1{}{};",
					vcut[0],
					acut[0],
					vcut[2],
					acut[2],
					channel('v', output),
					channel('a', output)
				);
			}
		}

		// add a newline after every filter to ease debugging
		writeln!(complex);
	}
}

pub(super) fn channel(channel: char, id: &str) -> String {
	if id.chars().any(|ch| !ch.is_digit(10)) {
		format!("[{channel}_{id}]")
	} else {
		format!("[{id}:{channel}]")
	}
}

fn next_tmp(filter_idx: &mut usize) -> String {
	*filter_idx += 1;
	format!("[tmp{filter_idx}]")
}

fn next_tmp_3(filter_idx: &mut usize) -> [String; 3] {
	[
		next_tmp(filter_idx),
		next_tmp(filter_idx),
		next_tmp(filter_idx)
	]
}
